"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.printFieldErrors = exports.validateFileExists = exports.getOSName = exports.findReadAndParseElmToolingJson = exports.elmToolingInstallPath = void 0;
const fs = require("fs");
const os = require("os");
const path = require("path");
const known_tools_1 = require("./known_tools");
const mixed_1 = require("./mixed");
const elmHome = process.env.ELM_HOME || path.join(os.homedir(), ".elm");
exports.elmToolingInstallPath = path.join(elmHome, "elm-tooling");
function findReadAndParseElmToolingJson() {
    const elmToolingJsonPath = mixed_1.findClosest("elm-tooling.json");
    if (elmToolingJsonPath === undefined) {
        return {
            tag: "ElmToolingJsonNotFound",
            message: "No elm-tooling.json found. To create one: elm-tooling init",
        };
    }
    let json = undefined;
    try {
        json = JSON.parse(fs.readFileSync(elmToolingJsonPath, "utf-8"));
    }
    catch (errorAny) {
        const error = errorAny;
        return {
            tag: "ReadAsJsonObjectError",
            elmToolingJsonPath,
            message: `Failed to read file as JSON:\n${error.message}`,
        };
    }
    if (!mixed_1.isRecord(json)) {
        return {
            tag: "ReadAsJsonObjectError",
            elmToolingJsonPath,
            message: `Expected an object but got: ${JSON.stringify(json)}`,
        };
    }
    const result = {
        tag: "Parsed",
        elmToolingJsonPath,
        unknownFields: [],
    };
    for (const [field, value] of Object.entries(json)) {
        switch (field) {
            case "entrypoints":
                result.entrypoints = prefixFieldResult("entrypoints", parseEntrypoints(elmToolingJsonPath, value));
                break;
            case "tools": {
                const osName = getOSName();
                result.tools = prefixFieldResult("tools", osName instanceof Error
                    ? {
                        tag: "Error",
                        errors: [
                            { path: [], message: osName.message },
                        ],
                    }
                    : parseTools(osName, value));
                break;
            }
            default:
                result.unknownFields.push(field);
                break;
        }
    }
    return result;
}
exports.findReadAndParseElmToolingJson = findReadAndParseElmToolingJson;
function getOSName() {
    switch (process.platform) {
        case "linux":
            return "linux";
        case "darwin":
            return "mac";
        case "win32":
            return "windows";
        default:
            return new Error(`Sorry, your platform (${process.platform}) is not supported yet :(`);
    }
}
exports.getOSName = getOSName;
function prefixFieldResult(prefix, fieldResult) {
    switch (fieldResult.tag) {
        case "Error":
            return {
                tag: "Error",
                errors: fieldResult.errors.map(({ path, message }) => ({
                    path: [prefix, ...path],
                    message,
                })),
            };
        case "Parsed":
            return fieldResult;
    }
}
function validateFileExists(fullPath) {
    try {
        const stats = fs.statSync(fullPath);
        if (!stats.isFile()) {
            return { tag: "Error", message: `Exists but is not a file: ${fullPath}` };
        }
    }
    catch (errorAny) {
        const error = errorAny;
        if (error.code === "ENOENT") {
            return {
                tag: "DoesNotExist",
                message: `File does not exist: ${fullPath}`,
            };
        }
        return {
            tag: "Error",
            message: `File error for ${fullPath}: ${error.message}`,
        };
    }
    return { tag: "Exists" };
}
exports.validateFileExists = validateFileExists;
function parseEntrypoints(elmToolingJsonPath, json) {
    if (!Array.isArray(json)) {
        return {
            tag: "Error",
            errors: [
                {
                    path: [],
                    message: `Expected an array but got: ${JSON.stringify(json)}`,
                },
            ],
        };
    }
    if (json.length === 0) {
        return {
            tag: "Error",
            errors: [
                {
                    path: [],
                    message: `Expected at least one entrypoint but got 0.`,
                },
            ],
        };
    }
    const [errors, entrypoints] = mixed_1.partitionMap(json, (entrypoint, index) => {
        if (typeof entrypoint !== "string") {
            return {
                tag: "Left",
                value: {
                    path: [index],
                    message: `Expected a string but got: ${JSON.stringify(entrypoint)}`,
                },
            };
        }
        if (!entrypoint.startsWith("./")) {
            return {
                tag: "Left",
                value: {
                    path: [index],
                    message: `Expected the string to start with "./" (to indicate that it is a relative path) but got: ${JSON.stringify(entrypoint)}`,
                },
            };
        }
        const absolutePath = path.join(path.dirname(elmToolingJsonPath), entrypoint);
        const exists = validateFileExists(absolutePath);
        if (exists.tag !== "Exists") {
            return {
                tag: "Left",
                value: { path: [index], message: exists.message },
            };
        }
        return {
            tag: "Right",
            value: {
                relativePath: entrypoint,
                absolutePath,
            },
        };
    });
    if (errors.length > 0) {
        return {
            tag: "Error",
            errors: errors,
        };
    }
    return { tag: "Parsed", parsed: entrypoints };
}
function parseTools(osName, json) {
    if (!mixed_1.isRecord(json)) {
        return {
            tag: "Error",
            errors: [
                {
                    path: [],
                    message: `Expected an object but got: ${JSON.stringify(json)}`,
                },
            ],
        };
    }
    const [errors, tools] = mixed_1.partitionMap(Object.entries(json), ([name, version]) => {
        if (typeof version !== "string") {
            return {
                tag: "Left",
                value: {
                    path: [name],
                    message: `Expected a version as a string but got: ${JSON.stringify(version)}`,
                },
            };
        }
        const versions = Object.prototype.hasOwnProperty.call(known_tools_1.KNOWN_TOOLS, name)
            ? known_tools_1.KNOWN_TOOLS[name]
            : undefined;
        if (versions === undefined) {
            return {
                tag: "Left",
                value: { path: [name], message: `Unknown tool` },
            };
        }
        const osAssets = Object.prototype.hasOwnProperty.call(versions, version)
            ? versions[version]
            : undefined;
        if (osAssets === undefined) {
            return {
                tag: "Left",
                value: { path: [name], message: `Unknown version: ${version}` },
            };
        }
        const tool = {
            name,
            version,
            absolutePath: getToolAbsolutePath(exports.elmToolingInstallPath, name, version),
            asset: osAssets[osName],
        };
        const exists = validateFileExists(tool.absolutePath);
        switch (exists.tag) {
            case "Exists":
                return {
                    tag: "Right",
                    value: [true, tool],
                };
            case "DoesNotExist":
                return {
                    tag: "Right",
                    value: [false, tool],
                };
            case "Error":
                return {
                    tag: "Left",
                    value: { path: [name], message: exists.message },
                };
        }
    });
    if (errors.length > 0) {
        return {
            tag: "Error",
            errors: errors,
        };
    }
    const [existing, missing] = mixed_1.partitionMap(tools, ([exists, tool]) => exists ? { tag: "Left", value: tool } : { tag: "Right", value: tool });
    return {
        tag: "Parsed",
        parsed: { existing, missing },
    };
}
function getToolAbsolutePath(elmToolingInstallPath, name, version) {
    return path.join(elmToolingInstallPath, name, version, name);
}
function printFieldErrors(errors) {
    return [
        mixed_1.printNumErrors(errors.length),
        ...errors.map((error) => `${mixed_1.bold(joinPath(error.path))}\n${mixed_1.indent(error.message)}`),
    ].join("\n\n");
}
exports.printFieldErrors = printFieldErrors;
function joinPath(errorPath) {
    if (errorPath.length === 0) {
        return "General";
    }
    const rest = errorPath
        .slice(1)
        .map((segment) => `[${JSON.stringify(segment)}]`);
    return `${errorPath[0]}${rest.join("")}`;
}
